@@ -28,20 +28,25 @@
 from __future__ import absolute_import, print_function
 
 import itertools
+from functools import partial
 
 from twisted.internet.defer import inlineCallbacks
 from twisted.internet.interfaces import IStreamClientEndpoint
 from twisted.internet.endpoints import UNIXClientEndpoint
 from twisted.internet.endpoints import TCP4ClientEndpoint
+from twisted.internet.error import ReactorNotRunning
 from twisted.internet.task import react
 
 try:
     _TLS = True
     from twisted.internet.endpoints import SSL4ClientEndpoint
     from twisted.internet.ssl import optionsForClientTLS, CertificateOptions
-    from twisted.internet.interfaces import IOpenSSLClientComponentCreator
-except ImportError:
+    from twisted.internet.interfaces import IOpenSSLClientConnectionCreator
+    from OpenSSL import SSL
+except ImportError as e:
     _TLS = False
+    if 'OpenSSL' not in str(e):
+        raise
 
 import txaio
 
@@ -52,16 +57,30 @@ from autobahn.wamp import component
 
 from autobahn.twisted.util import sleep
 from autobahn.twisted.wamp import ApplicationSession
+from autobahn.wamp.exception import ApplicationError
 
 
 __all__ = ('Component')
 
 
+def _is_ssl_error(e):
+    """
+    Internal helper.
+
+    This is so we can just return False if we didn't import any
+    TLS/SSL libraries. Otherwise, returns True if this is an
+    OpenSSL.SSL.Error
+    """
+    if _TLS:
+        return isinstance(e, SSL.Error)
+    return False
+
+
 def _unique_list(seq):
     """
     Return a list with unique elements from sequence, preserving order.
     """
-    seen = set(seq)
+    seen = set()
     return [x for x in seq if x not in seen and not seen.add(x)]
 
 
@@ -98,7 +117,7 @@ def _create_transport_serializers(transport_config):
     Create a list of serializers to use with a WAMP protocol factory.
     """
     if u'serializers' in transport_config:
-        serializer_ids = _unique_list(transport_config['serializers'])
+        serializer_ids = _unique_list(transport_config[u'serializers'])
     else:
         serializer_ids = [u'msgpack', u'json']
 
@@ -115,7 +134,7 @@ def _create_transport_serializers(transport_config):
                 serializers.append(MsgPackSerializer(batched=True))
                 serializers.append(MsgPackSerializer())
 
-        if serializer_id == u'json':
+        elif serializer_id == u'json':
             # try JSON WAMP serializer
             try:
                 from autobahn.wamp.serializer import JsonSerializer
@@ -125,6 +144,11 @@ def _create_transport_serializers(transport_config):
                 serializers.append(JsonSerializer(batched=True))
                 serializers.append(JsonSerializer())
 
+        else:
+            raise RuntimeError(
+                "Unknown serializer '{}'".format(serializer_id)
+            )
+
     return serializers
 
 
@@ -168,9 +192,9 @@ def _create_transport_endpoint(reactor, endpoint_config):
                     raise RuntimeError('TLS configured in transport, but TLS support is not installed (eg OpenSSL?)')
 
                 # FIXME: create TLS context from configuration
-                if IOpenSSLClientComponentCreator.providedBy(tls):
+                if IOpenSSLClientConnectionCreator.providedBy(tls):
                     # eg created from twisted.internet.ssl.optionsForClientTLS()
-                    context = IOpenSSLClientComponentCreator(tls)
+                    context = IOpenSSLClientConnectionCreator(tls)
 
                 elif isinstance(tls, CertificateOptions):
                     context = tls
@@ -225,11 +249,34 @@ class Component(component.Component):
 
     log = txaio.make_logger()
 
-    session = ApplicationSession
+    session_factory = ApplicationSession
     """
     The factory of the session we will instantiate.
     """
 
+    def _check_native_endpoint(self, endpoint):
+        if IStreamClientEndpoint.providedBy(endpoint):
+            pass
+        elif isinstance(endpoint, dict):
+            if 'tls' in endpoint:
+                tls = endpoint['tls']
+                if isinstance(tls, (dict, bool)):
+                    pass
+                elif IOpenSSLClientConnectionCreator.providedBy(tls):
+                    pass
+                elif isinstance(tls, CertificateOptions):
+                    pass
+                else:
+                    raise ValueError(
+                        "'tls' configuration must be a dict, CertificateOptions or"
+                        " IOpenSSLClientConnectionCreator provider"
+                    )
+        else:
+            raise ValueError(
+                "'endpoint' configuration must be a dict or IStreamClientEndpoint"
+                " provider"
+            )
+
     def _connect_transport(self, reactor, transport_config, session_factory):
         """
         Create and connect a WAMP-over-XXX transport.
@@ -238,16 +285,27 @@ class Component(component.Component):
         transport_endpoint = _create_transport_endpoint(reactor, transport_config['endpoint'])
         return transport_endpoint.connect(transport_factory)
 
+    # XXX think: is it okay to use inlineCallbacks (in this
+    # twisted-only file) even though we're using txaio?
     @inlineCallbacks
     def start(self, reactor=None):
+        """
+        This starts the Component, which means it will start connecting
+        (and re-connecting) to its configured transports. A Component
+        runs until it is "done", which means one of:
+
+        - There was a "main" function defined, and it completed successfully;
+        - Something called ``.leave()`` on our session, and we left successfully;
+        - ``.stop()`` was called, and completed successfully;
+        - none of our transports were able to connect successfully (failure);
+
+        :returns: a Deferred that fires (with ``None``) when we are
+            "done" or with a Failure if something went wrong.
+        """
         if reactor is None:
+            self.log.warn("Using default reactor")
             from twisted.internet import reactor
 
-        # txaio.use_twisted()
-        # txaio.config.loop = reactor
-
-        # txaio.start_logging(level='debug')
-
         yield self.fire('start', reactor, self)
 
         # transports to try again and again ..
@@ -255,7 +313,7 @@ class Component(component.Component):
 
         reconnect = True
 
-        self.log.info('entering recomponent loop')
+        self.log.debug('Entering re-connect loop')
 
         while reconnect:
             # cycle through all transports forever ..
@@ -272,18 +330,51 @@ class Component(component.Component):
                 )
                 yield sleep(delay)
                 try:
-                    transport.connect_attempts += 1
-                    yield self._connect_once(reactor, transport.config)
-                    transport.connect_sucesses += 1
+                    yield self._connect_once(reactor, transport)
+                    self.log.info('Component completed successfully')
+
                 except Exception as e:
-                    transport.connect_failures += 1
-                    self.log.error(u'component failed: {error}', error=e)
+                    # need concept of "fatal errors", for which a
+                    # connection is *never* going to work. Might want
+                    # to also add, for example, things like
+                    # SyntaxError
+                    if isinstance(e, ApplicationError):
+                        if e.error in [u'wamp.error.no_such_realm']:
+                            reconnect = False
+                            self.log.error(u"Fatal error, not reconnecting")
+                            raise
+                        # self.log.error(u"{error}: {message}", error=e.error, message=e.message)
+                    elif _is_ssl_error(e):
+                        # Quoting pyOpenSSL docs: "Whenever
+                        # [SSL.Error] is raised directly, it has a
+                        # list of error messages from the OpenSSL
+                        # error queue, where each item is a tuple
+                        # (lib, function, reason). Here lib, function
+                        # and reason are all strings, describing where
+                        # and what the problem is. See err(3) for more
+                        # information."
+                        for (lib, fn, reason) in e.args[0]:
+                            self.log.error(u"TLS failure: {reason}", reason=reason)
+                        self.log.error(u"Marking this transport as failed")
+                        transport.failed()
+                    else:
+                        f = txaio.create_failure()
+                        self.log.error(
+                            u'Connection failed: {error}',
+                            error=txaio.failure_message(f),
+                        )
+                        # some types of errors should probably have
+                        # stacktraces logged immediately at error
+                        # level, e.g. SyntaxError?
+                        self.log.debug(u'{tb}', tb=txaio.failure_format_traceback(f))
+                        raise
                 else:
                     reconnect = False
             else:
                 # check if there is any transport left we can use
                 # to connect
                 if not self._can_reconnect():
+                    self.log.info("No remaining transports to try")
                     reconnect = False
 
 
@@ -292,19 +383,46 @@ def _run(reactor, components):
         components = [components]
 
     if type(components) != list:
-        raise RuntimeError('"components" must be a list of Component objects - encountered {0}'.format(type(components)))
+        raise ValueError(
+            '"components" must be a list of Component objects - encountered'
+            ' {0}'.format(type(components))
+        )
 
     for c in components:
         if not isinstance(c, Component):
-            raise RuntimeError('"components" must be a list of Component objects - encountered item of type {0}'.format(type(c)))
+            raise ValueError(
+                '"components" must be a list of Component objects - encountered'
+                'item of type {0}'.format(type(c))
+            )
+
+    log = txaio.make_logger()
+
+    def component_success(c, arg):
+        log.debug("Component {c} successfully completed: {arg}", c=c, arg=arg)
+        return arg
+
+    def component_failure(f):
+        log.error("Component error: {msg}", msg=txaio.failure_message(f))
+        log.debug("Component error: {tb}", tb=txaio.failure_format_traceback(f))
+        return None
 
     # all components are started in parallel
     dl = []
     for c in components:
         # a component can be of type MAIN or SETUP
-        dl.append(c.start(reactor))
+        d = c.start(reactor)
+        txaio.add_callbacks(d, partial(component_success, c), component_failure)
+        dl.append(d)
+    d = txaio.gather(dl, consume_exceptions=False)
 
-    d = txaio.gather(dl, consume_exceptions=True)
+    def all_done(arg):
+        log.debug("All components ended; stopping reactor")
+        try:
+            reactor.stop()
+        except ReactorNotRunning:
+            pass
+
+    txaio.add_callbacks(d, all_done, all_done)
 
     return d
 
