@@ -26,6 +26,8 @@
 
 from __future__ import absolute_import
 
+import txaio
+
 from twisted.internet.protocol import Factory
 from twisted.protocols.basic import Int32StringReceiver
 from twisted.internet.error import ConnectionDone
@@ -34,8 +36,6 @@ from autobahn.twisted.util import peer2str, transport_channel_id
 from autobahn.util import _LazyHexFormatter
 from autobahn.wamp.exception import ProtocolError, SerializationError, TransportLost
 
-import txaio
-
 __all__ = (
     'WampRawSocketServerProtocol',
     'WampRawSocketClientProtocol',
@@ -63,6 +63,9 @@ class WampRawSocketProtocol(Int32StringReceiver):
         else:
             self.peer = peer2str(peer)
 
+        # a Future/Deferred that fires when we hit STATE_CLOSED
+        self.is_closed = txaio.create_future()
+
         # this will hold an ApplicationSession object
         # once the RawSocket opening handshake has been
         # completed
@@ -94,10 +97,11 @@ class WampRawSocketProtocol(Int32StringReceiver):
             self.log.warn("WampRawSocketProtocol: ApplicationSession constructor / onOpen raised ({err})", err=e)
             self.abort()
         else:
-            self.log.info("ApplicationSession started.")
+            self.log.debug("ApplicationSession started.")
 
     def connectionLost(self, reason):
-        self.log.info("WampRawSocketProtocol: connection lost: reason = '{reason}'", reason=reason)
+        self.log.debug("WampRawSocketProtocol: connection lost: reason = '{reason}'", reason=reason)
+        txaio.resolve(self.is_closed, self)
         try:
             wasClean = isinstance(reason.value, ConnectionDone)
             self._session.onClose(wasClean)
@@ -107,10 +111,10 @@ class WampRawSocketProtocol(Int32StringReceiver):
         self._session = None
 
     def stringReceived(self, payload):
-        self.log.debug("WampRawSocketProtocol: RX octets: {octets}", octets=_LazyHexFormatter(payload))
+        self.log.trace("WampRawSocketProtocol: RX octets: {octets}", octets=_LazyHexFormatter(payload))
         try:
             for msg in self._serializer.unserialize(payload):
-                self.log.debug("WampRawSocketProtocol: RX WAMP message: {msg}", msg=msg)
+                self.log.trace("WampRawSocketProtocol: RX WAMP message: {msg}", msg=msg)
                 self._session.onMessage(msg)
 
         except ProtocolError as e:
@@ -126,7 +130,7 @@ class WampRawSocketProtocol(Int32StringReceiver):
         Implements :func:`autobahn.wamp.interfaces.ITransport.send`
         """
         if self.isOpen():
-            self.log.debug("WampRawSocketProtocol: TX WAMP message: {msg}", msg=msg)
+            self.log.trace("WampRawSocketProtocol: TX WAMP message: {msg}", msg=msg)
             try:
                 payload, _ = self._serializer.serialize(msg)
             except Exception as e:
@@ -134,7 +138,7 @@ class WampRawSocketProtocol(Int32StringReceiver):
                 raise SerializationError("WampRawSocketProtocol: unable to serialize WAMP application payload ({0})".format(e))
             else:
                 self.sendString(payload)
-                self.log.debug("WampRawSocketProtocol: TX octets: {octets}", octets=_LazyHexFormatter(payload))
+                self.log.trace("WampRawSocketProtocol: TX octets: {octets}", octets=_LazyHexFormatter(payload))
         else:
             raise TransportLost()
 
@@ -361,8 +365,10 @@ class WampRawSocketServerFactory(WampRawSocketFactory):
            :class:`autobahn.wamp.interfaces.ISerializer`.
         :type serializers: list
         """
-        assert(callable(factory))
-        self._factory = factory
+        if callable(factory):
+            self._factory = factory
+        else:
+            self._factory = lambda: factory
 
         if serializers is None:
             serializers = []
@@ -424,8 +430,10 @@ class WampRawSocketClientFactory(WampRawSocketFactory):
            :class:`autobahn.wamp.interfaces.ISerializer`.
         :type serializer: obj
         """
-        assert(callable(factory))
-        self._factory = factory
+        if callable(factory):
+            self._factory = factory
+        else:
+            self._factory = lambda: factory
 
         if serializer is None:
 
