@@ -39,12 +39,14 @@ from autobahn.wamp.types import Challenge
 
 from twisted.internet.defer import inlineCallbacks, returnValue
 
-__all__ = []
+__all__ = [
+    'HAS_CRYPTOSIGN',
+    'HAS_CRYPTOSIGN_SSHAGENT'
+]
 
 try:
-    # try to import everything that is optional for Autobahn in
-    # general, but which we need for WAMP-cryptosign
-    from nacl import public, encoding, signing
+    # try to import everything we need for WAMP-cryptosign
+    from nacl import public, encoding, signing, bindings
 except ImportError:
     HAS_CRYPTOSIGN = False
     HAS_CRYPTOSIGN_SSHAGENT = False
@@ -95,13 +97,17 @@ def _pack(keyparts):
 
 def _read_ssh_ed25519_pubkey(keydata):
     """
-    Parse a Ed25519 SSH public key from a string into a raw public key.
+    Parse an OpenSSH Ed25519 public key from a string into a raw public key.
 
-    :param keydata: The public Ed25519 SSH key to parse.
+    Example input:
+
+        ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIJukDU5fqXv/yVhSirsDWsUFyOodZyCSLxyitPPzWJW9 oberstet@office-corei7
+
+    :param keydata: The OpenSSH Ed25519 public key data to parse.
     :type keydata: unicode
 
-    :returns: The raw public keys (32 bytes).
-    :rtype: binary
+    :returns: pair of raw public key (32 bytes) and comment
+    :rtype: tuple
     """
     if type(keydata) != six.text_type:
         raise Exception("invalid type {} for keydata".format(type(keydata)))
@@ -124,7 +130,233 @@ def _read_ssh_ed25519_pubkey(keydata):
     if len(key) != 32:
         raise Exception('invalid length {} for embedded raw key (must be 32 bytes)'.format(len(key)))
 
-    return key
+    return key, comment
+
+
+class _SSHPacketReader:
+    """
+    Read OpenSSH packet format which is used for key material.
+    """
+
+    def __init__(self, packet):
+        self._packet = packet
+        self._idx = 0
+        self._len = len(packet)
+
+    def get_remaining_payload(self):
+        return self._packet[self._idx:]
+
+    def get_bytes(self, size):
+        if self._idx + size > self._len:
+            raise Exception('incomplete packet')
+
+        value = self._packet[self._idx:self._idx + size]
+        self._idx += size
+        return value
+
+    def get_uint32(self):
+        return int.from_bytes(self.get_bytes(4), 'big')
+
+    def get_string(self):
+        return self.get_bytes(self.get_uint32())
+
+
+def _read_ssh_ed25519_privkey(keydata):
+    """
+    Parse an OpenSSH Ed25519 private key from a string into a raw private key.
+
+    Example input:
+
+        -----BEGIN OPENSSH PRIVATE KEY-----
+        b3BlbnNzaC1rZXktdjEAAAAABG5vbmUAAAAEbm9uZQAAAAAAAAABAAAAMwAAAAtzc2gtZW
+        QyNTUxOQAAACCbpA1OX6l7/8lYUoq7A1rFBcjqHWcgki8corTz81iVvQAAAKDWjZ0Y1o2d
+        GAAAAAtzc2gtZWQyNTUxOQAAACCbpA1OX6l7/8lYUoq7A1rFBcjqHWcgki8corTz81iVvQ
+        AAAEArodzIMjH9MOBz0X+HDvL06rEJOMYFhzGQ5zXPM7b7fZukDU5fqXv/yVhSirsDWsUF
+        yOodZyCSLxyitPPzWJW9AAAAFm9iZXJzdGV0QG9mZmljZS1jb3JlaTcBAgMEBQYH
+        -----END OPENSSH PRIVATE KEY-----
+
+
+    :param keydata: The OpenSSH Ed25519 private key data to parse.
+    :type keydata: unicode
+
+    :returns: pair of raw private key (32 bytes) and comment
+    :rtype: tuple
+    """
+
+    # Some pointers:
+    # https://github.com/ronf/asyncssh/blob/master/asyncssh/public_key.py
+    # https://github.com/ronf/asyncssh/blob/master/asyncssh/ed25519.py
+    # crypto_sign_ed25519_sk_to_seed
+    # https://github.com/jedisct1/libsodium/blob/master/src/libsodium/crypto_sign/ed25519/sign_ed25519_api.c#L27
+    # https://tools.ietf.org/html/draft-bjh21-ssh-ed25519-02
+    # http://blog.oddbit.com/2011/05/08/converting-openssh-public-keys/
+
+    SSH_BEGIN = u'-----BEGIN OPENSSH PRIVATE KEY-----'
+    SSH_END = u'-----END OPENSSH PRIVATE KEY-----'
+    OPENSSH_KEY_V1 = b'openssh-key-v1\0'
+
+    if not (keydata.startswith(SSH_BEGIN) and keydata.endswith(SSH_END)):
+        raise Exception('invalid OpenSSH private key (does not start/end with OPENSSH preamble)')
+
+    ssh_end = keydata.find(SSH_END)
+    keydata = keydata[len(SSH_BEGIN):ssh_end]
+    keydata = u''.join([x.strip() for x in keydata.split()])
+    blob = binascii.a2b_base64(keydata)
+
+    blob = blob[len(OPENSSH_KEY_V1):]
+    packet = _SSHPacketReader(blob)
+
+    cipher_name = packet.get_string()
+    kdf = packet.get_string()
+    packet.get_string()  # kdf_data
+    nkeys = packet.get_uint32()
+    packet.get_string()  # public_key
+    key_data = packet.get_string()
+    mac = packet.get_remaining_payload()
+
+    block_size = 8
+
+    if cipher_name != b'none':
+        raise Exception('encrypted private keys not supported (please remove the passphrase from your private key or use SSH agent)')
+
+    if kdf != b'none':
+        raise Exception('passphrase encrypted private keys not supported')
+
+    if nkeys != 1:
+        raise Exception('multiple private keys in a key file not supported (found {} keys)'.format(nkeys))
+
+    if mac:
+        raise Exception('invalid OpenSSH private key (found remaining payload for mac)')
+
+    packet = _SSHPacketReader(key_data)
+
+    packet.get_uint32()  # check1
+    packet.get_uint32()  # check2
+
+    alg = packet.get_string()
+
+    if alg != b'ssh-ed25519':
+        raise Exception('invalid key type: we only support Ed25519 (found "{}")'.format(alg.decode('ascii')))
+
+    vk = packet.get_string()
+    sk = packet.get_string()
+
+    if len(vk) != bindings.crypto_sign_PUBLICKEYBYTES:
+        raise Exception('invalid public key length')
+
+    if len(sk) != bindings.crypto_sign_SECRETKEYBYTES:
+        raise Exception('invalid public key length')
+
+    comment = packet.get_string()                             # comment
+    pad = packet.get_remaining_payload()
+
+    if len(pad) >= block_size or pad != bytes(range(1, len(pad) + 1)):
+        raise Exception('invalid OpenSSH private key')
+
+    # secret key (64 octets) = 32 octets seed || 32 octets secret key derived of seed
+    seed = sk[:bindings.crypto_sign_SEEDBYTES]
+
+    comment = comment.decode('ascii')
+
+    return seed, comment
+
+
+def _read_signify_ed25519_signature(signature_file):
+    """
+    Read a Ed25519 signature file created with OpenBSD signify.
+
+    http://man.openbsd.org/OpenBSD-current/man1/signify.1
+    """
+    with open(signature_file) as f:
+        # signature file format: 2nd line is base64 of 'Ed' || 8 random octets || 64 octets Ed25519 signature
+        sig = binascii.a2b_base64(f.read().splitlines()[1])[10:]
+        if len(sig) != 64:
+            raise Exception('bogus Ed25519 signature: raw signature length was {}, but expected 64'.format(len(sig)))
+        return sig
+
+
+def _read_signify_ed25519_pubkey(pubkey_file):
+    """
+    Read a public key from a Ed25519 key pair created with OpenBSD signify.
+
+    http://man.openbsd.org/OpenBSD-current/man1/signify.1
+    """
+    with open(pubkey_file) as f:
+        # signature file format: 2nd line is base64 of 'Ed' || 8 random octets || 32 octets Ed25519 public key
+        pubkey = binascii.a2b_base64(f.read().splitlines()[1])[10:]
+        if len(pubkey) != 32:
+            raise Exception('bogus Ed25519 public key: raw key length was {}, but expected 32'.format(len(pubkey)))
+        return pubkey
+
+
+def _qrcode_from_signify_ed25519_pubkey(pubkey_file, mode='text'):
+    """
+
+    Usage:
+
+    1. Get the OpenBSD 5.7 release public key from here
+
+        http://cvsweb.openbsd.org/cgi-bin/cvsweb/src/etc/signify/Attic/openbsd-57-base.pub?rev=1.1
+
+    2. Generate QR Code and print to terminal
+
+        print(cryptosign._qrcode_from_signify_ed25519_pubkey('openbsd-57-base.pub'))
+
+    3. Compare to (scroll down) QR code here
+
+        https://www.openbsd.org/papers/bsdcan-signify.html
+    """
+    assert(mode in ['text', 'svg'])
+
+    import pyqrcode
+
+    with open(pubkey_file) as f:
+        pubkey = f.read().splitlines()[1]
+
+        qr = pyqrcode.create(pubkey, error='L', mode='binary')
+
+        if mode == 'text':
+            return qr.terminal()
+
+        elif mode == 'svg':
+            import io
+            data_buffer = io.BytesIO()
+
+            qr.svg(data_buffer, omithw=True)
+
+            return data_buffer.getvalue()
+
+        else:
+            raise Exception('logic error')
+
+
+def _verify_signify_ed25519_signature(pubkey_file, signature_file, message):
+    """
+    Verify a Ed25519 signature created with OpenBSD signify.
+
+    This will raise a `nacl.exceptions.BadSignatureError` if the signature is bad
+    and return silently when the signature is good.
+
+    Usage:
+
+    1. Create a signature:
+
+        signify-openbsd -S -s ~/.signify/crossbario-trustroot.sec -m .profile
+
+    2. Verify the signature
+
+        from autobahn.wamp import cryptosign
+
+        with open('.profile', 'rb') as f:
+            message = f.read()
+            cryptosign._verify_signify_ed25519_signature('.signify/crossbario-trustroot.pub', '.profile.sig', message)
+
+    http://man.openbsd.org/OpenBSD-current/man1/signify.1
+    """
+    pubkey = _read_signify_ed25519_pubkey(pubkey_file)
+    verify_key = signing.VerifyKey(pubkey)
+    sig = _read_signify_ed25519_signature(signature_file)
+    verify_key.verify(message, sig)
 
 
 # SigningKey from
@@ -305,29 +537,21 @@ if HAS_CRYPTOSIGN:
             key (from a SSH private key file) or a (public) verification key (from a SSH
             public key file). A private key file must be passphrase-less.
             """
-            # https://tools.ietf.org/html/draft-bjh21-ssh-ed25519-02
-            # http://blog.oddbit.com/2011/05/08/converting-openssh-public-keys/
-
             SSH_BEGIN = u'-----BEGIN OPENSSH PRIVATE KEY-----'
-            SSH_END = u'-----END OPENSSH PRIVATE KEY-----'
 
             with open(filename, 'r') as f:
                 keydata = f.read().strip()
 
-            if keydata.startswith(SSH_BEGIN) and keydata.endswith(SSH_END):
-                # SSH private key
-                # ssh_end = keydata.find(SSH_END)
-                # keydata = keydata[len(SSH_BEGIN):ssh_end]
-                # keydata = u''.join([x.strip() for x in keydata.split()])
-                # blob = binascii.a2b_base64(keydata)
-                # prefix = 'openssh-key-v1\x00'
-                # data = unpack(blob[len(prefix):])
-                raise Exception("loading private keys not implemented")
+            if keydata.startswith(SSH_BEGIN):
+                # OpenSSH private key
+                keydata, comment = _read_ssh_ed25519_privkey(keydata)
+                key = signing.SigningKey(keydata, encoder=encoding.RawEncoder)
             else:
-                # SSH public key
-                keydata = _read_ssh_ed25519_pubkey(filename)
-                key = public.PublicKey(keydata)
-                return cls(key)
+                # OpenSSH public key
+                keydata, comment = _read_ssh_ed25519_pubkey(filename)
+                key = public.PublicKey(keydata, encoder=encoding.RawEncoder)
+
+            return cls(key, comment)
 
 
 if HAS_CRYPTOSIGN_SSHAGENT:
@@ -359,7 +583,7 @@ if HAS_CRYPTOSIGN_SSHAGENT:
             if not HAS_CRYPTOSIGN_SSHAGENT:
                 raise Exception("SSH agent integration is not supported on this platform")
 
-            pubkey = _read_ssh_ed25519_pubkey(pubkey)
+            pubkey, _ = _read_ssh_ed25519_pubkey(pubkey)
 
             if not reactor:
                 from twisted.internet import reactor
