@@ -31,6 +31,7 @@ import txaio
 import inspect
 
 from autobahn import wamp
+from autobahn.util import IdGenerator, ObservableMixin
 from autobahn.wamp import uri
 from autobahn.wamp import message
 from autobahn.wamp import types
@@ -39,9 +40,7 @@ from autobahn.wamp import exception
 from autobahn.wamp.exception import ApplicationError, ProtocolError, SessionNotReady, SerializationError
 from autobahn.wamp.interfaces import IApplicationSession  # noqa
 from autobahn.wamp.types import SessionDetails
-from autobahn.wamp.keyring import EncryptedPayload
-from autobahn.util import IdGenerator, ObservableMixin
-
+from autobahn.wamp.cryptobox import EncryptedPayload
 from autobahn.wamp.request import \
     Publication, \
     Subscription, \
@@ -72,15 +71,15 @@ class BaseSession(ObservableMixin):
         """
 
         """
-        ObservableMixin.__init__(self)
-
-        # this is for library level debugging
-        self.debug = False
-
-        # this is for app level debugging. exceptions raised in user code
-        # will get logged (that is, when invoking remoted procedures or
-        # when invoking event handlers)
-        self.debug_app = False
+        self.set_valid_events(
+            valid_events=[
+                'join',         # right before onJoin runs
+                'leave',        # after onLeave has run
+                'ready',        # after onJoin and all 'join' listeners have completed
+                'connect',      # right before onConnect
+                'disconnect',   # right after onDisconnect
+            ]
+        )
 
         # this is for marshalling traceback from exceptions thrown in user
         # code within WAMP error messages (that is, when invoking remoted
@@ -204,14 +203,24 @@ class BaseSession(ObservableMixin):
                     encrypted_payload = EncryptedPayload(msg.enc_algo, msg.enc_key, msg.enc_serializer, msg.payload)
                     decrypted_error, msg.args, msg.kwargs = self._keyring.decrypt(True, msg.error, encrypted_payload)
                 except Exception as e:
-                    log_msg = u"failed to decrypt application payload 1: {}".format(e)
-                    self.log.warn(log_msg)
-                    enc_err = ApplicationError(ApplicationError.ENC_DECRYPT_ERROR, log_msg, enc_algo=msg.enc_algo)
+                    self.log.warn("failed to decrypt application payload 1: {err}", err=e)
+                    enc_err = ApplicationError(
+                        ApplicationError.ENC_DECRYPT_ERROR,
+                        u"failed to decrypt application payload 1: {}".format(e),
+                        enc_algo=msg.enc_algo,
+                    )
                 else:
                     if msg.error != decrypted_error:
-                        log_msg = u"URI within encrypted payload ('{}') does not match the envelope ('{}')".format(decrypted_error, msg.error)
-                        self.log.warn(log_msg)
-                        enc_err = ApplicationError(ApplicationError.ENC_TRUSTED_URI_MISMATCH, log_msg, enc_algo=msg.enc_algo)
+                        self.log.warn(
+                            u"URI within encrypted payload ('{decrypted_error}') does not match the envelope ('{error}')",
+                            decrypted_error=decrypted_error,
+                            error=msg.error,
+                        )
+                        enc_err = ApplicationError(
+                            ApplicationError.ENC_TRUSTED_URI_MISMATCH,
+                            u"URI within encrypted payload ('{}') does not match the envelope ('{}')".format(decrypted_error, msg.error),
+                            enc_algo=msg.enc_algo,
+                        )
 
         if enc_err:
             return enc_err
@@ -308,11 +317,16 @@ class ApplicationSession(BaseSession):
         Implements :func:`autobahn.wamp.interfaces.ITransportHandler.onOpen`
         """
         self._transport = transport
-        d = txaio.as_future(self.onConnect)
-
-        def _error(e):
-            return self._swallow_error(e, "While firing onConnect")
-        txaio.add_callbacks(d, None, _error)
+        d = self.fire('connect', self, transport)
+        txaio.add_callbacks(
+            d, None,
+            lambda fail: self._swallow_error(fail, "While notifying 'connect'")
+        )
+        txaio.add_callbacks(
+            d,
+            lambda _: txaio.as_future(self.onConnect),
+            None,
+        )
 
     def onConnect(self):
         """
@@ -363,7 +377,7 @@ class ApplicationSession(BaseSession):
         """
         Implements :func:`autobahn.wamp.interfaces.ISession.is_attached`
         """
-        return self._session_id is not None
+        return self._transport is not None and self._session_id is not None
 
     def onUserError(self, fail, msg):
         """
@@ -382,7 +396,9 @@ class ApplicationSession(BaseSession):
             suggested you log this immediately after the exception.
         """
         if isinstance(fail.value, exception.ApplicationError):
-            self.log.error(fail.value.error_message())
+            # silence on errors raised explicitly from the app
+            # previous code: self.log.error(fail.value.error_message())
+            pass
         else:
             self.log.error(
                 u'{msg}: {traceback}',
@@ -404,14 +420,18 @@ class ApplicationSession(BaseSession):
         '''
         try:
             self.onUserError(fail, msg)
-        except:
-            pass
+        except Exception:
+            self.log.error(
+                "Internal error: {tb}",
+                tb=txaio.failure_format_traceback(txaio.create_failure()),
+            )
         return None
 
     def onMessage(self, msg):
         """
         Implements :func:`autobahn.wamp.interfaces.ITransportHandler.onMessage`
         """
+
         if self._session_id is None:
 
             # the first message must be WELCOME, ABORT or CHALLENGE ..
@@ -423,11 +443,40 @@ class ApplicationSession(BaseSession):
                 self._session_id = msg.session
 
                 details = SessionDetails(self._realm, self._session_id, msg.authid, msg.authrole, msg.authmethod, msg.authprovider, msg.authextra)
-                d = txaio.as_future(self.onJoin, details)
-
-                def _error(e):
-                    return self._swallow_error(e, "While firing onJoin")
-                txaio.add_callbacks(d, None, _error)
+                # firing 'join' *before* running onJoin, so that the
+                # idiom where you "do stuff" in onJoin -- possibly
+                # including self.leave() -- works properly. Besides,
+                # there's "ready" that fires after 'join' and onJoin
+                # have all completed...
+                d = self.fire('join', self, details)
+                # add a logging errback first, which will ignore any
+                # errors from fire()
+                txaio.add_callbacks(
+                    d, None,
+                    lambda e: self._swallow_error(e, "While notifying 'join'")
+                )
+                # this should run regardless
+                txaio.add_callbacks(
+                    d,
+                    lambda _: txaio.as_future(self.onJoin, details),
+                    None
+                )
+                # ignore any errors from onJoin (XXX or, should that be fatal?)
+                txaio.add_callbacks(
+                    d, None,
+                    lambda e: self._swallow_error(e, "While firing onJoin")
+                )
+                # this instance is now "ready"...
+                txaio.add_callbacks(
+                    d,
+                    lambda _: self.fire('ready', self),
+                    None
+                )
+                # ignore any errors from 'ready'
+                txaio.add_callbacks(
+                    d, None,
+                    lambda e: self._swallow_error(e, "While notifying 'ready'")
+                )
 
             elif isinstance(msg, message.Abort):
 
@@ -435,9 +484,14 @@ class ApplicationSession(BaseSession):
                 details = types.CloseDetails(msg.reason, msg.message)
                 d = txaio.as_future(self.onLeave, details)
 
+                def success(arg):
+                    # XXX also: handle async
+                    self.fire('leave', self, details)
+                    return arg
+
                 def _error(e):
                     return self._swallow_error(e, "While firing onLeave")
-                txaio.add_callbacks(d, None, _error)
+                txaio.add_callbacks(d, success, _error)
 
             elif isinstance(msg, message.Challenge):
 
@@ -462,9 +516,14 @@ class ApplicationSession(BaseSession):
                     details = types.CloseDetails(reply.reason, reply.message)
                     d = txaio.as_future(self.onLeave, details)
 
+                    def success(arg):
+                        # XXX also: handle async
+                        self.fire('leave', self, details)
+                        return arg
+
                     def _error(e):
                         return self._swallow_error(e, "While firing onLeave")
-                    txaio.add_callbacks(d, None, _error)
+                    txaio.add_callbacks(d, success, _error)
                     # switching to the callback chain, effectively
                     # cancelling error (which we've now handled)
                     return d
@@ -488,10 +547,15 @@ class ApplicationSession(BaseSession):
                 details = types.CloseDetails(msg.reason, msg.message)
                 d = txaio.as_future(self.onLeave, details)
 
+                def success(arg):
+                    # XXX also: handle async
+                    self.fire('leave', self, details)
+                    return arg
+
                 def _error(e):
                     errmsg = 'While firing onLeave for reason "{0}" and message "{1}"'.format(msg.reason, msg.message)
                     return self._swallow_error(e, errmsg)
-                txaio.add_callbacks(d, None, _error)
+                txaio.add_callbacks(d, success, _error)
 
             elif isinstance(msg, message.Event):
 
@@ -523,7 +587,7 @@ class ApplicationSession(BaseSession):
                         invoke_kwargs = msg.kwargs if msg.kwargs else dict()
 
                         if handler.details_arg:
-                            invoke_kwargs[handler.details_arg] = types.EventDetails(publication=msg.publication, publisher=msg.publisher, topic=topic, enc_algo=msg.enc_algo)
+                            invoke_kwargs[handler.details_arg] = types.EventDetails(publication=msg.publication, publisher=msg.publisher, publisher_authid=msg.publisher_authid, publisher_authrole=msg.publisher_authrole, topic=topic, enc_algo=msg.enc_algo)
 
                         def _error(e):
                             errmsg = 'While firing {0} subscribed under {1}.'.format(
@@ -609,14 +673,25 @@ class ApplicationSession(BaseSession):
                                 encrypted_payload = EncryptedPayload(msg.enc_algo, msg.enc_key, msg.enc_serializer, msg.payload)
                                 decrypted_proc, msg.args, msg.kwargs = self._keyring.decrypt(True, proc, encrypted_payload)
                             except Exception as e:
-                                log_msg = u"failed to decrypt application payload 1: {}".format(e)
-                                self.log.warn(log_msg)
-                                enc_err = ApplicationError(ApplicationError.ENC_DECRYPT_ERROR, log_msg)
+                                self.log.warn(
+                                    "failed to decrypt application payload 1: {err}",
+                                    err=e,
+                                )
+                                enc_err = ApplicationError(
+                                    ApplicationError.ENC_DECRYPT_ERROR,
+                                    u"failed to decrypt application payload 1: {}".format(e),
+                                )
                             else:
                                 if proc != decrypted_proc:
-                                    log_msg = u"URI within encrypted payload ('{}') does not match the envelope ('{}')".format(decrypted_proc, proc)
-                                    self.log.warn(log_msg)
-                                    enc_err = ApplicationError(ApplicationError.ENC_TRUSTED_URI_MISMATCH, log_msg)
+                                    self.log.warn(
+                                        "URI within encrypted payload ('{decrypted_proc}') does not match the envelope ('{proc}')",
+                                        decrypted_proc=decrypted_proc,
+                                        proc=proc,
+                                    )
+                                    enc_err = ApplicationError(
+                                        ApplicationError.ENC_TRUSTED_URI_MISMATCH,
+                                        u"URI within encrypted payload ('{}') does not match the envelope ('{}')".format(decrypted_proc, proc),
+                                    )
 
                     if msg.progress:
                         # process progressive call result
@@ -695,14 +770,26 @@ class ApplicationSession(BaseSession):
                                     encrypted_payload = EncryptedPayload(msg.enc_algo, msg.enc_key, msg.enc_serializer, msg.payload)
                                     decrypted_proc, msg.args, msg.kwargs = self._keyring.decrypt(False, proc, encrypted_payload)
                                 except Exception as e:
-                                    log_msg = u"failed to decrypt INVOCATION payload: {}".format(e)
-                                    self.log.warn(log_msg)
-                                    enc_err = ApplicationError(ApplicationError.ENC_DECRYPT_ERROR, log_msg)
+                                    self.log.warn(
+                                        "failed to decrypt INVOCATION payload: {err}",
+                                        err=e,
+                                    )
+                                    enc_err = ApplicationError(
+                                        ApplicationError.ENC_DECRYPT_ERROR,
+                                        "failed to decrypt INVOCATION payload: {}".format(e),
+                                    )
                                 else:
                                     if proc != decrypted_proc:
-                                        log_msg = u"URI within encrypted INVOCATION payload ('{}') does not match the envelope ('{}')".format(decrypted_proc, proc)
-                                        self.log.warn(log_msg)
-                                        enc_err = ApplicationError(ApplicationError.ENC_TRUSTED_URI_MISMATCH, log_msg)
+                                        self.log.warn(
+                                            "URI within encrypted INVOCATION payload ('{decrypted_proc}') "
+                                            "does not match the envelope ('{proc}')",
+                                            decrypted_proc=decrypted_proc,
+                                            proc=proc,
+                                        )
+                                        enc_err = ApplicationError(
+                                            ApplicationError.ENC_TRUSTED_URI_MISMATCH,
+                                            u"URI within encrypted INVOCATION payload ('{}') does not match the envelope ('{}')".format(decrypted_proc, proc),
+                                        )
 
                         if enc_err:
                             # when there was a problem decrypting the INVOCATION payload, we obviously can't invoke
@@ -750,7 +837,7 @@ class ApplicationSession(BaseSession):
                                 else:
                                     progress = None
 
-                                invoke_kwargs[endpoint.details_arg] = types.CallDetails(progress, caller=msg.caller, procedure=proc, enc_algo=msg.enc_algo)
+                                invoke_kwargs[endpoint.details_arg] = types.CallDetails(progress, caller=msg.caller, caller_authid=msg.caller_authid, caller_authrole=msg.caller_authrole, procedure=proc, enc_algo=msg.enc_algo)
 
                             on_reply = txaio.as_future(endpoint.fn, *invoke_args, **invoke_kwargs)
 
@@ -769,8 +856,10 @@ class ApplicationSession(BaseSession):
                                             else:
                                                 encrypted_payload = self._keyring.encrypt(False, proc, [res])
                                         except Exception as e:
-                                            log_msg = u"failed to encrypt application payload: {}".format(e)
-                                            self.log.warn(log_msg)
+                                            self.log.warn(
+                                                "failed to encrypt application payload: {err}",
+                                                err=e,
+                                            )
 
                                 if encrypted_payload:
                                     reply = message.Yield(msg.request,
@@ -933,16 +1022,30 @@ class ApplicationSession(BaseSession):
 
         if self._session_id:
             # fire callback and close the transport
-            d = txaio.as_future(self.onLeave, types.CloseDetails(reason=types.CloseDetails.REASON_TRANSPORT_LOST, message="WAMP transport was lost without closing the session before"))
+            details = types.CloseDetails(
+                reason=types.CloseDetails.REASON_TRANSPORT_LOST,
+                message=(u"WAMP transport was lost without closing the"
+                         u" session before"),
+            )
+            d = txaio.as_future(self.onLeave, details)
+
+            def success(arg):
+                # XXX also: handle async
+                self.fire('leave', self, details)
+                return arg
 
             def _error(e):
                 return self._swallow_error(e, "While firing onLeave")
-            txaio.add_callbacks(d, None, _error)
+            txaio.add_callbacks(d, success, _error)
 
             self._session_id = None
 
         d = txaio.as_future(self.onDisconnect)
 
+        def success(arg):
+            # XXX do we care about returning 'arg' properly?
+            return self.fire('disconnect', self, was_clean=wasClean)
+
         def _error(e):
             return self._swallow_error(e, "While firing onDisconnect")
         txaio.add_callbacks(d, None, _error)
@@ -957,7 +1060,6 @@ class ApplicationSession(BaseSession):
         """
         Implements :func:`autobahn.wamp.interfaces.ISession.onJoin`
         """
-        return self.fire('join', self, details)
 
     def onLeave(self, details):
         """
@@ -966,8 +1068,6 @@ class ApplicationSession(BaseSession):
         if details.reason.startswith('wamp.error.'):
             self.log.error('{reason}: {wamp_message}', reason=details.reason, wamp_message=details.message)
 
-        self.fire('leave', self, details)
-
         if self._transport:
             self.disconnect()
         # do we ever call onLeave with a valid transport?
@@ -995,7 +1095,7 @@ class ApplicationSession(BaseSession):
         """
         Implements :func:`autobahn.wamp.interfaces.ISession.onDisconnect`
         """
-        return self.fire('disconnect', self, True)
+        pass  # return self.fire('disconnect', self, True)
 
     def publish(self, topic, *args, **kwargs):
         """
